// Scilab 2.4 script

// WARNING THIS SCRIPT DOES NOT RUN UNDER SCILAB 2.6

// BAM associative memory for a microthesaurus

clear;

// number of bits on which the letters are encoded
b = 5;
// maximal size of "x" and "y" strings: no. of letters
NL = 10;
KL =10;
N = NL * b;
K = KL * b;

// list of words, there is crosstalk when working in reverse
printf("Word List",1)
v = ['constant  ', 'invariable';
     'emotional ', 'heartfelt ';
     'hereditary', 'ancestral ';
     'intolerant', 'dogmatic  ';
     'miniature ', 'diminutive';
     'regretful ', 'apologetic';
     'squeamish ', 'fastidious']

P = size(v,'r');


//--------------------------------------------------------------------
// convert to bipolar vectors

// initialize x and y
x = zeros(N,P);
y = zeros(K,P);


// first binary
code_x = zeros(N);
code_y = zeros(K);
for p = 1 : P
  code_x = str2code(v(p,1)) - str2code('a');
  // first duplicate
  x(:,p) = code_x .*. ones(b,1);
  for i = 0 : NL - 1
    x(i*b+1,p) = int(x(i*b+1,p)/16);
    x(i*b+2,p) = int((x(i*b+2,p) - 16*x(i*b+1,p))/8);
    x(i*b+3,p) = int((x(i*b+3,p) - 16*x(i*b+1,p) - 8*x(i*b+2,p))/4);
    x(i*b+4,p) = int((x(i*b+4,p) - 16*x(i*b+1,p) - 8*x(i*b+2,p) ...
                                 - 4*x(i*b+3,p))/2);
    x(i*b+5,p) = modulo(x(i*b+5,p),2);
  end;
  // same as for "x"
  code_y = str2code(v(p,2)) - str2code('a');
  y(:,p) = code_y .*. ones(b,1);
  for i = 0 : KL - 1
    y(i*b+1,p) = int(y(i*b+1,p)/16);
    y(i*b+2,p) = int((y(i*b+2,p) - 16*y(i*b+1,p))/8);
    y(i*b+3,p) = int((y(i*b+3,p) - 16*y(i*b+1,p) - 8*y(i*b+2,p))/4);
    y(i*b+4,p) = int((y(i*b+4,p) - 16*y(i*b+1,p) - 8*y(i*b+2,p) ...
                                 - 4*y(i*b+3,p))/2);
    y(i*b+5,p) = modulo(y(i*b+5,p),2);
  end;
end;

// transform to bipolar vectors
x = 2 * x - ones(N,P);
y = 2 * y - ones(K,P);

//--------------------------------------------------------------------
// initialize and calculate weights
W = zeros(K,N);
for p = 1 : P
  W = W + y(:,p) * x(:,p)';
end;


//--------------------------------------------------------------------
// test forward

// init
x_test = zeros(N,1);
y_test = zeros(K,1);
x_out = - ones(NL,1);
y_out = - ones(KL,1);
direct_test = emptystr(P,2);
indirect_test = emptystr(P,2);

// do
for p = 1 : P

  // --- direct run ---
  // initialize
  x_test = x(:,p);
  y_test = W * x_test;

  // find the vectors
  for t = 1 : 30
    A = sign(W' * y_test);
    x_test = A + (ones(N,1) - abs(A)) .* x_test;
    B = sign(W * x_test);
    y_test = B + (ones(K,1) - abs(B)) .* y_test;
  end;

  // convert x_test, y_test back to binary
  x_test = (x_test + ones(N,1))/2;
  y_test = (y_test + ones(K,1))/2;

  // convert to letter codes
  for il = 1 : NL
    x_out(il) = x_test((il-1) * b + 1) * 16 ...
                + x_test((il-1) * b + 2) * 8 ...
                + x_test((il-1) * b + 3) * 4 ...
                + x_test((il-1) * b + 4) * 2 ...
                + x_test((il-1) * b + 5) ...
                + str2code('a');
  end;
  // convert to letter codes
  for il = 1 : KL
    y_out(il) = y_test((il-1) * b + 1) * 16 ...
                + y_test((il-1) * b + 2) * 8 ...
                + y_test((il-1) * b + 3) * 4 ...
                + y_test((il-1) * b + 4) * 2 ...
                + y_test((il-1) * b + 5) ...
                + str2code('a');
  end;

  // store
  direct_test(p,1) = code2str(x_out);
  direct_test(p,2) = code2str(y_out);

  // --- reverse run ---
  // initialize
  y_test = y(:,p);
  x_test = W' * y_test;

  // find the vectors
  for t = 1 : 30
    A = sign(W * x_test);
    y_test = A + (ones(K,1) - abs(A)) .* y_test;
    B = sign(W' * y_test);
    x_test = B + (ones(N,1) - abs(B)) .* x_test;
  end;

  // convert x_test, y_test back to binary
  x_test = (x_test + ones(N,1))/2;
  y_test = (y_test + ones(K,1))/2;

  // convert to letter codes
  for il = 1 : NL
    x_out(il) = x_test((il-1) * b + 1) * 16 ...
                + x_test((il-1) * b + 2) * 8 ...
                + x_test((il-1) * b + 3) * 4 ...
                + x_test((il-1) * b + 4) * 2 ...
                + x_test((il-1) * b + 5) ...
                + str2code('a');
  end;
  // convert to letter codes
  for il = 1 : KL
    y_out(il) = y_test((il-1) * b + 1) * 16 ...
                + y_test((il-1) * b + 2) * 8 ...
                + y_test((il-1) * b + 3) * 4 ...
                + y_test((il-1) * b + 4) * 2 ...
                + y_test((il-1) * b + 5) ...
                + str2code('a');
  end;

  // store
  indirect_test(p,1) = code2str(x_out);
  indirect_test(p,2) = code2str(y_out);
end;

// printout
printf("Direct run results",1)
direct_test
printf("Indirect run results",1)
indirect_test

// find bit difference between vectors

// convert back to binary vectors
x = (x + ones(N,P))/2;
y = (y + ones(K,P))/2;

// initialize
x_bit_diff = zeros(P,P);
y_bit_diff = zeros(P,P);

//do
for p1 = 1 : P
  for p2 = 1 : P
    x_bit_diff(p1,p2) = sum(abs(x(:,p1) - x(:,p2)));
    y_bit_diff(p1,p2) = sum(abs(y(:,p1) - y(:,p2)));
  end;
end;

// printout
printf("Bit differences between x vectors",1)
x_bit_diff
printf("Bit differences between y vectors",1)
y_bit_diff
