// Scilab 2.4 script

// WARNING THIS SCRIPT DOES NOT RUN UNDER SCILAB 2.6

// Counter propagation network to recognize letters A, B, C 
// represented as a matrix 5x6 points

/////////////////////////////////////////////////////////////////////
// Data

// Input representative vectors:
// A =   *   = 00100   B = ****  = 11110  C =  **** = 01111
//      * *    01010       *   *   10001      *       10000
//     *   *   10001       ****    11110      *       10000
//     *****   11111       *   *   10001      *       10000
//     *   *   10001       *   *   10001      *       10000
//     *   *   10001       ****    11110       ****   01111
//
// The return value is the ASCII code of the letter
// A = 65 = 01000001  B = 66 = 01000010  C = 67 = 01000011
//
// the network have 38=30+8 input/output neurons corresponding to input vectors
// and 3 hidden neurons wich will classify input x vectors
// in one of 3 classes: A, B, C

// Representative vectors, one for each class/hidden neuron:
// rep_x matrix, containing representative x vectors (input) on rows
// one (row) for each class/hidden neuron <=> no of rows = h
rep_x=[0,0,1,0,0, 0,1,0,1,0, 1,0,0,0,1, 1,1,1,1,1, 1,0,0,0,1, 1,0,0,0,1; ...
       1,1,1,1,0, 1,0,0,0,1, 1,1,1,1,0, 1,0,0,0,1, 1,0,0,0,1, 1,1,1,1,0; ...
       0,1,1,1,1, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 0,1,1,1,1]';

// rep_y matrix, containing representative y vectors (ASCII code) on rows
// one (row) for each class/hidden neuron <=> no of rows = h
rep_y=[0,1,0,0,0,0,0,1; ...
       0,1,0,0,0,0,1,0; ...
       0,1,0,0,0,0,1,1]';

//-------------------------------------------------------------------
// Input training vectors:
// first  set have one missing "*"
// second set have one supplemental "*"
// third  set have both
// A =       = 00000 ;   *   = 00100 ;   *   = 00100
//      * *    01010    * *    01010    * *    01010
//     *   *   10001   * * *   10101   * * *   10101
//     *****   11111   *****   11111   *****   11111
//     *   *   10001   *   *   10001   *   *   10001
//     *   *   10001   *   *   10001       *   00001
//
// B = ****  = 11110 ; ****  = 11110 ; ****  = 11110
//         *   00001   *   *   10001   *   *   10001
//     ****    11110   *****   11111   *****   11111
//     *   *   10001   *   *   10001       *   00001
//     *   *   10001   *   *   10001   *   *   10001
//     ****    11110   ****    11110   ****    11110
//
// C =  **** = 01111 ; ***** = 11111 ; ***** = 11111
//             00000   *       10000   *       10000
//     *       10000   *       10000   *       10000
//     *       10000   *       10000   *       10000
//     *       10000   *       10000   *       10000
//      ****   01111    ****   01111    ***    01110

// x matrix, containing x vectors (input) on rows
train_x=[0,0,0,0,0, 0,1,0,1,0, 1,0,0,0,1, 1,1,1,1,1, 1,0,0,0,1, 1,0,0,0,1; ...
         0,0,1,0,0, 0,1,0,1,0, 1,0,1,0,1, 1,1,1,1,1, 1,0,0,0,1, 1,0,0,0,1; ...
         0,0,1,0,0, 0,1,0,1,0, 1,0,1,0,1, 1,1,1,1,1, 1,0,0,0,1, 0,0,0,0,1; ...
         1,1,1,1,0, 0,0,0,0,1, 1,1,1,1,0, 1,0,0,0,1, 1,0,0,0,1, 1,1,1,1,0; ...
         1,1,1,1,0, 1,0,0,0,1, 1,1,1,1,1, 1,0,0,0,1, 1,0,0,0,1, 1,1,1,1,0; ...
         1,1,1,1,0, 1,0,0,0,1, 1,1,1,1,1, 0,0,0,0,1, 1,0,0,0,1, 1,1,1,1,0; ...
         0,1,1,1,1, 0,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 0,1,1,1,1; ...
         1,1,1,1,1, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 0,1,1,1,1; ...
         1,1,1,1,1, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 0,1,1,1,0]';

// y matrix, containing y vectors (ASCII code) on rows
train_y=[0,1,0,0,0,0,0,1; ...
         0,1,0,0,0,0,0,1; ...
         0,1,0,0,0,0,0,1; ...
         0,1,0,0,0,0,1,0; ...
         0,1,0,0,0,0,1,0; ...
         0,1,0,0,0,0,1,0; ...
         0,1,0,0,0,0,1,1; ...
         0,1,0,0,0,0,1,1; ...
         0,1,0,0,0,0,1,1]';

//-------------------------------------------------------------------
// characters to be recognized, used at run time (after learning ended)
// first  set have one missing "*"
// second set have one supplemental "*"
// third  set have both
// fourth set is an extreme
// they are different from training set
//
// NOTE: Characters B and C have many common "*", 
// they may be easily mixed up
//
// A =   *   = 00100 ;   *   = 00100 ;   *   = 00100 ;   *   = 00100
//      * *    01010    * *    01010    * *    01010           00000
//     *   *   10001   *   *   10001   *   *   10001   *       10000
//     *****   11111   *****   11111   *****   11111      **   00011
//     *   *   10001   * * *   10101   * * *   10101           00000
//     *       10000   *   *   10001   *       10000   *   *   10001
//
// B = ****  = 11110 ; ****  = 11110 ; ****  = 11110 ; *  *  = 10010
//     *   *   10001   *   *   10001   *   *   10001           00000
//     ****    11110   ****    11110   ****    11110   * **    10110
//     *   *   10001   *   *   10001   *   *   10001           00000
//     *       10000   * * *   10101   * *     10100           00000
//     ****    11110   ****    11110   ****    11110   *  *    10010
//
// C =  * ** = 01011 ;  **** = 01111 ;  * ** = 01011 ;   **  = 00110
//     *       10000   *       10000   *       10000           00000
//     *       10000   *       10000   *       10000   *       10000
//     *       10000   *       10000   *       10000   *       10000
//     *       10000   **      11000   **      11000           00000
//      ****   01111    ****   01111    ****   01111     **    00110

// x matrix, containing x vectors (input) on rows
// order on rows is: A, B, C,  A, C, B,  B, A, C,  C, B, A (some randomness)
run_x=[0,0,1,0,0, 0,1,0,1,0, 1,0,0,0,1, 1,1,1,1,1, 1,0,0,0,1, 1,0,0,0,0; ...
       1,1,1,1,0, 1,0,0,0,1, 1,1,1,1,0, 1,0,0,0,1, 1,0,0,0,0, 1,1,1,1,0; ...
       0,1,0,1,1, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 0,1,1,1,1; ...
       0,0,1,0,0, 0,1,0,1,0, 1,0,0,0,1, 1,1,1,1,1, 1,0,1,0,1, 1,0,0,0,1; ...
       0,1,1,1,1, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 1,1,0,0,0, 0,1,1,1,1; ...
       1,1,1,1,0, 1,0,0,0,1, 1,1,1,1,1, 1,0,0,0,1, 1,0,1,0,1, 1,1,1,1,0; ...
       1,1,1,1,0, 1,0,0,0,1, 1,1,1,1,0, 1,0,0,0,1, 1,0,1,0,0, 1,1,1,1,0; ...
       0,0,1,0,0, 0,1,0,1,0, 1,0,0,0,1, 1,1,1,1,1, 1,0,1,0,1, 1,0,0,0,0; ...
       0,1,0,1,1, 1,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 1,1,0,0,0, 0,1,1,1,1; ...
       0,0,1,1,0, 0,0,0,0,0, 1,0,0,0,0, 1,0,0,0,0, 0,0,0,0,0, 0,0,1,1,0; ...
       1,0,0,1,0, 0,0,0,0,0, 1,0,1,1,0, 0,0,0,0,0, 0,0,0,0,0, 1,0,0,1,0; ...
       0,0,1,0,0, 0,0,0,0,0, 1,0,0,0,0, 0,0,0,1,1, 0,0,0,0,0, 1,0,0,0,1]';

// run_y is implicitly initialized with 0

//-------------------------------------------------------------------
// Miscelanneous variables

// dimension of x vector and input/output x layer
dim_x = size(rep_x,'r');

// dimension of y vector and input/output y layer
dim_y = size(rep_y,'r');

// dimension of hidden layer/no of classes: 3 letters => 3 classes
dim_hidden = size(rep_x,'c');

// no. of training vectors <=> no. of columns in train_x & train_y matrices
no_training_vec = size(train_x,'c');

// no. of running vectors <=> no. of columns in run_x 
no_running_vec = size(run_x,'c');

// no of training cicles
// for hidden layer
no_train_cycles_hidden = 1;
// and for output layer
no_train_cycles_out = 1;

//-------------------------------------------------------------------
// Learning constants

// for hidden layer
c = 0.1;

// for output layer
c1 = 0.1;

/////////////////////////////////////////////////////////////////////
// Normalize the input vectors <=> implement input level functionality

// initialize the storage for normalized input vectors
train_z = zeros(dim_x + dim_y, no_training_vec);

for i = 1 : no_training_vec,
  // norm of input vector
  train_norm = sqrt(sum(train_x(:,i) .* train_x(:,i)) ...
                    + sum(train_y(:,i) .* train_y(:,i)));
  
  // build normalized vectors
  train_z(1 : dim_x, i) = train_x(:,i) / train_norm;
  train_z(dim_x + 1 : dim_x + dim_y, i) = train_y(:,i) / train_norm;
end;

/////////////////////////////////////////////////////////////////////
// Learning procedure

// initialize the storage for weights
// will store a normalized input representative vector per row
W = zeros(dim_hidden, dim_x + dim_y);

// initialize weights on hidden layer with normalized representative vectors
for i=1 : dim_hidden,
  // first normalize, analogous previous procedure
  // (it is not worthwile to write a special procedure for normalization)
  rep_norm = sqrt(sum(rep_x(:,i) .* rep_x(:,i)) ...
                      + sum(rep_y(:,i) .* rep_y(:,i)));
  W(i, 1 : dim_x) = rep_x(:,i)' / rep_norm;
  W(i, dim_x + 1 : dim_x + dim_y) = rep_y(:,i)' / rep_norm;
end;

//-------------------------------------------------------------------
// train hidden layer with given training vectors

// the hidden layer is trained trough the full set of data 
// for "no_train_cycles_hidden" times

for i = 1 : no_train_cycles_hidden,
  // find winner
  [win_val,winner] = max(W * train_z(:,i));
  // update weights
  W(winner,:) = W(winner,:) + c * (train_z(:,i)' - W(winner,:));
end;

//-------------------------------------------------------------------
// initialize weights on output layer with representative vectors

// initialize weight storage on output layer
W1 = zeros(dim_x + dim_y, dim_hidden);

for i=1 : dim_hidden,
  // initialize with representatives
  W1(1:dim_x,i) = rep_x(:,i);
  W1(dim_x+1 : dim_x+dim_y, i) = rep_y(:,i);
end;

//-------------------------------------------------------------------
// train output layer

// the output layer is trained trough the full set of data 
// for "no_train_cycles_out" times

for i = 1 : no_train_cycles_out,
  // find winner on hidden layer
  [win_val,winner] = max(W * train_z(:,i));
  // update winner weights, "x" and "y" parts
  W1(1 : dim_x, winner) = W1(1 : dim_x, winner) ...
                          + c1 * (train_x(:,i) - W1(1 : dim_x, winner));
  W1(dim_x + 1 : dim_x + dim_y, winner) = ...
                          W1(dim_x + 1 : dim_x + dim_y, winner) ...
                          + c1 * (train_y(:,i) ...
                                  - W1(dim_x + 1 : dim_x + dim_y, winner));
end;

/////////////////////////////////////////////////////////////////////
// running procedure
// x vector will be applied (letter form)
// and y vector will be retrieved (ASCII code)

// Normalize the input vectors <=> implement input level functionality
// initialize the storage for normalized input vectors
run_z = zeros(dim_x + dim_y, no_running_vec);
for i = 1 : no_running_vec,
  // y vector is made 0 here
  run_norm = sqrt(sum(run_x(:,i) .* run_x(:,i)));
  run_z(1 : dim_x, i) = run_x(:,i) / run_norm;
  // run_z, y part remains 0
end;

// initialize storage for y output vector
y_out = zeros(dim_y, no_running_vec);

for i=1 : no_running_vec,
  // find the winning hidden neuron
  [win_val,winner] = max(W * run_z(:,i));

  // y elements should be either 0 or 1, so use round
  y_out(:,i) = round(W1(dim_x+1 : dim_x+dim_y,winner));
end;

/////////////////////////////////////////////////////////////////////
// convert from binary matrix to  ASCII codes

for i=1 : no_running_vec,
  ASCII_code = 0;
  for j = dim_y - 1 : -1 : 0,
    // y1 elements should be 0 or 1 so use round
    ASCII_code = ASCII_code + 2^j * y_out(dim_y-j, i);
  end;
  // in SciLab majuscules start at -10 downwards !
  printf('The character no. %i is: %c', i, 55 - ASCII_code);
end;
