// SOM/Kohonen map K neurons arranged in a square LxL layout (K = L^2)
// mapping a square

/////////////////////////////////////////////////////////////////////
// Depending over constants chosen and due to the fact that the
// learning set is randomly chosen, the learning may not be always
// succesfull.
/////////////////////////////////////////////////////////////////////

// this script make some snapshots from learning process

/////////////////////////////////////////////////////////////////////
// Variables and setup

clear;

// open figure drawing
driver('Fig');
current_driver = driver();
xinit('som_w_map.fig');
// figure1 is at t=0
// figure1 is at t=5
figure2 = 6;
// figure1 is at t=70
figure3 = 71;
// figure4 is at t=T

// starting point for random numbers always the same (for comparison purposes)
rand('seed',0);

// number of neurons:
//   per side
L = 8;
//   per total
K = L^2;

// T total number of time steps
T = 50000;

// size of input vectors
N = 2;

// initialize data, 
// total number of patterns equals total number of time steps
x = rand(T,N);

// initialize weights
W = rand(K,N);

// stop function
tau = 1;
tau_rate = 0.9999;

// distance const. in lateral feedback function
d = 4^2;
d_rate = 0.99995^2;
// feedback height
h = 0.8;

// this vector helps compute distances
delta = linspace(1,L,L)';

// plot style
plotstyle = ones(1,L);

// train net and draw figure
for t = 1 : T;
  // --- figure drawing ---

  if t == 1 | t == figure2 | t == figure3 then
    // draw figure for t-1 (weights are from previous cicle)
    // set window
    select t
      // first figure for t = 0, upper left
      case 1    then win_left = 0;   win_up = 0; ,
      // second figure for t = fig2-1, upper right
      case figure2 then win_left = 0.5; win_up = 0; ,
      // third figure for t = fig3-1, lower left
      case figure3 then win_left = 0;   win_up = 0.5; ,
    end;

    xsetech([win_left,win_up,0.5,0.5]);
    
    // write title (time)
    xtitle('t = ' + string(t - 1));
    // W_x holds the "x" coordinates of neuron's weight vectors in weight space
    W_x = matrix(W(:,1),L,L)';
    // W_x holds the "y" coordinates of neuron's weight vectors in weight space
    W_y = matrix(W(:,2),L,L)';
    // plot matrix row horizontal lines
    plot2d(W_x, W_y, plotstyle, '011', "x@y", [0,0,1,1], [1,10,1,10]);
    // plot matrix column vertical lines
    plot2d(W_x' ,W_y', plotstyle, "000");
  end;
    
  // --- network training ---
  
  // - winner first -
  // find distances
  // first differences of type [x(t) - weight_x, y(t) - weight_y]
  XY = (ones(K,1) * x(t,:)) - W;
  // and then squared Euclidean distances
  dist = sum(XY .* XY, 'c');
  // and winner k
  [winner_dist,k] = min(dist);
  // k : x coordinate
  k_x = modulo(k-1,L) + 1;
  // k : y coordinate
  k_y = (k - k_x)/L + 1;
    
  // - stop function -
  tau = tau * tau_rate;
  
  // - lateral feedback function -
  d = d * d_rate;
  
  // "x" and "y" set of squared distances (reduced by current d)
  dd_x = delta - k_x * ones(L,1);
  d_x = dd_x .* dd_x / d;
  dd_y = delta - k_y * ones(L,1);
  d_y = dd_y .* dd_y / d;
    
  // lateral feedback (note the order of multiplication!)
  psi = h * (exp(- d_y) .*. exp(- d_x));
    
  // update weights
  W = W + (tau * psi * ones(1,N)) .* (ones(K,1) * x(t,:) - W);
end;


// draw last snapshot
win_left = 0.5; win_up = 0.5;
xsetech([win_left,win_up,0.5,0.5]);

// write title (time)
xtitle('t = ' + string(50000));
// W_x holds the "x" coordinates of neuron's weight vectors in weight space
W_x = matrix(W(:,1),L,L)';
// W_x holds the "y" coordinates of neuron's weight vectors in weight space
W_y = matrix(W(:,2),L,L)';
// plot matrix row horizontal lines
plot2d(W_x, W_y, plotstyle, '011', "x@y", [0,0,1,1], [1,10,1,10]);
// plot matrix column vertical lines
plot2d(W_x' ,W_y', plotstyle, "000");
