// SOM/Kohonen map K neurons arranged in a square LxL layout (K = L^2)
// mapping a square


/////////////////////////////////////////////////////////////////////
// Depending over constants chosen and due to the fact that the
// learning set is randomly chosen, the learning may not be always
// succesfull.
/////////////////////////////////////////////////////////////////////

// animation of network learning, showed in weights space,
// to be run from inside SciLab

/////////////////////////////////////////////////////////////////////
// Variables and setup

clear;

// starting point for random numbers always the same (for comparison purposes)
rand('seed',0);

// number of neurons:
//   per side
L = 8;
//   per total
K = L^2;

// T total number of time steps
T = 50000;

// size of input vectors
N = 2;

// initialize data, 
// total number of patterns equals total number of time steps
x = rand(T,N);

// initialize weights
W = rand(K,N);

// stop function
tau = 1;
tau_rate = 0.9999;

// distance const. in lateral feedback function
d = 4^2;
d_rate = 0.99995^2;
// feedback height
h = 0.8;

// this vector helps compute distances
delta = linspace(1,L,L)';

// plot style
plotstyle = ones(1,L);

// train net and draw figure

// first 100 sequences
for tt = 1 : 100
  // --- figure drawing ---
  // clear window (old figure)
  xbasc();
  // write title (time)
  xtitle('t = ' + string(tt - 1));
  // W_x holds the "x" coordinates of neuron's weight vectors in weight space
  W_x = matrix(W(:,1),L,L)';
  // W_x holds the "y" coordinates of neuron's weight vectors in weight space
  W_y = matrix(W(:,2),L,L)';
  // plot matrix row horizontal lines
  plot2d(W_x, W_y, plotstyle, '011', "x@y", [0,0,1,1], [1,10,1,10]);
  // plot matrix column vertical lines
  plot2d(W_x' ,W_y', plotstyle, "000");

  // make a pause if necessary
  // unix_s("sleep 1s");

  // --- network training ---
  
  // - winner first -
  // find distances
  // first differences of type [x(t) - weight_x, y(t) - weight_y]
  XY = (ones(K,1) * x(tt,:)) - W;
  // and then squared Euclidean distances
  dist = sum(XY .* XY, 'c');
  // and winner k
  [winner_dist,k] = min(dist);
  // k : x coordinate
  k_x = modulo(k-1,L) + 1;
  // k : y coordinate
  k_y = (k - k_x)/L + 1;
  
  // - stop function -
  tau = tau * tau_rate;
  
  // - lateral feedback function -
  d = d * d_rate;
  
  // "x" and "y" set of squared distances (reduced by current d)
  dd_x = delta - k_x * ones(L,1);
  d_x = dd_x .* dd_x / d;
  dd_y = delta - k_y * ones(L,1);
  d_y = dd_y .* dd_y / d;
  
  // lateral feedback (note the order of multiplication!)
  psi = h * (exp(- d_y) .*. exp(- d_x));
  
  // update weights
  W = W + (tau * psi * ones(1,N)) .* (ones(K,1) * x(tt,:) - W);
end;

// next, in 100 step size
for t = 1 : T/100 - 1;
  // --- figure drawing ---
  // clear window (old figure)
  xbasc();
  // write title (time)
  xtitle('t = ' + string(100 * t));
  // W_x holds the "x" coordinates of neuron's weight vectors in weight space
  W_x = matrix(W(:,1),L,L)';
  // W_x holds the "y" coordinates of neuron's weight vectors in weight space
  W_y = matrix(W(:,2),L,L)';
  // plot matrix row horizontal lines
  plot2d(W_x, W_y, plotstyle, '011', "x@y", [0,0,1,1], [1,10,1,10]);
  // plot matrix column vertical lines
  plot2d(W_x' ,W_y', plotstyle, "000");
  
  for tt = 1 : 100
    // --- network training ---
    
    // - winner first -
    // find distances
    // first differences of type [x(t) - weight_x, y(t) - weight_y]
    XY = (ones(K,1) * x(100*t + tt,:)) - W;
    // and then squared Euclidean distances
    dist = sum(XY .* XY, 'c');
    // and winner k
    [winner_dist,k] = min(dist);
    // k : x coordinate
    k_x = modulo(k-1,L) + 1;
    // k : y coordinate
    k_y = (k - k_x)/L + 1;
    // - stop function -
    tau = tau * tau_rate;
    
    // - lateral feedback function -
    d = d * d_rate;
    
    // "x" and "y" set of squared distances (reduced by current d)
    dd_x = delta - k_x * ones(L,1);
    d_x = dd_x .* dd_x / d;
    dd_y = delta - k_y * ones(L,1);
    d_y = dd_y .* dd_y / d;
    
    // lateral feedback (note the order of multiplication!)
    psi = h * (exp(- d_y) .*. exp(- d_x));
    
    // update weights
    W = W + (tau * psi * ones(1,N)) .* (ones(K,1) * x(100*t + tt,:) - W);
  end;
end;

// --- last figure drawing ---
// clear window (old figure)
xbasc();
// write title (time)
xtitle('t = ' + string(T));
// W_x holds the "x" coordinates of neuron's weight vectors in weight space
W_x = matrix(W(:,1),L,L)';
// W_x holds the "y" coordinates of neuron's weight vectors in weight space
W_y = matrix(W(:,2),L,L)';
// plot matrix row horizontal lines
plot2d(W_x, W_y, plotstyle, '011', "x@y", [0,0,1,1], [1,10,1,10]);
// plot matrix column vertical lines
plot2d(W_x' ,W_y', plotstyle, "000");
