// SOM/Kohonen map: K neurons arranged in a line 
// mapping a square

/////////////////////////////////////////////////////////////////////
// Depending over constants chosen and due to the fact that the
// learning set is randomly chosen, the learning may not be always
// succesfull.
/////////////////////////////////////////////////////////////////////

// this script takes some snapshots from a linear network
// learning a bidimensional space

/////////////////////////////////////////////////////////////////////
// Variables and setup

clear;

// open figure drawing
driver('Fig');
current_driver = driver();
xinit('som_w_map_lin.fig');

// starting point for random numbers always the same (for comparison purposes)
rand('seed',0);

// number of neurons:
K = 64;

// T total number of time steps
T = 50000;

// size of input vectors
N = 2;

// initialize data, 
// total number of patterns equals total number of time steps
x = rand(T,N);

// initialize weights
W = rand(K,N);

// stop function
tau = 1;
// tau_rate = 0.999;
tau_rate = 0.9999;

// distance const. in lateral feedback function
d = 15^2;
d_rate = 0.9999^2;
// feedback height
h = 0.8;

// this vector helps compute distances
delta = linspace(1,K,K)';

// plot style
plotstyle = ones(1,K);

// train net and draw figure
for t = 1 : T;
  // --- figure drawing ---
  if t == 8001 then
    // set upper-left window
    xsetech([0,0,0.5,0.5]);
    // write title (time)
    xtitle('t = ' + string(t - 1));
    // plot lines between weights
    plot2d(W(:,1), W(:,2), plotstyle, '011', "x@y", [0,0,1,1], [1,10,1,10]);
    // plot weights as small circles
    plot2d(W(:,1) ,W(:,2), - 9 * plotstyle, "000");
  end;
    
  // --- network training ---
  
  // - winner first -
  // find distances
  // first differences of type [x(t) - weight_x, y(t) - weight_y]
  XY = (ones(K,1) * x(t,:)) - W;
  // and then squared Euclidean distances
  dist = sum(XY .* XY, 'c');
  // and winner k
  [winner_dist,k] = min(dist);
  
  // - stop function -
  tau = tau * tau_rate;
  
  // - lateral feedback function -
  d = d * d_rate;
  
  // "x" set of squared distances (reduced by current d)
  dd_x = delta - k * ones(K,1);
  d_x = dd_x .* dd_x / d;
  
  // lateral feedback (note the order of multiplication!)
  psi = h * exp(- d_x);
  
  // update weights
  W = W + (tau * psi * ones(1,N)) .* (ones(K,1) * x(t,:) - W);
end;

// set upper-left window
xsetech([0.5,0,0.5,0.5]);
// write title (time)
xtitle('t = ' + string(50000));
// plot lines between weights
plot2d(W(:,1), W(:,2), plotstyle, '011', "x@y", [0,0,1,1], [1,10,1,10]);
// plot weights as small circles
plot2d(W(:,1) ,W(:,2), - 9 * plotstyle, "000");
